# resources/gcg/__init__.py

# SPDX-FileCopyrightText: Portions Copyright (c) 2023 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

# Greedy Coordinate Gradient implementation lightly modified from https://github.com/llm-attacks/llm-attacks
# Paper can be found at: https://arxiv.org/abs/2307.15043

from .generate_gcg import gcg_parser, run_gcg
