<!DOCTYPE html>
<html lang="en">

<head>
<meta name="viewport" content="width=device-width, initial-scale=1">
<meta charset="UTF-8" />
<style>
body {font-family: sans-serif}
.defcon1 {background-color: #f94144; text-color: #000}
.defcon2 {background-color: #f8961e; text-color: #000}
.defcon3 {background-color: #ccc; text-color: #000}
.defcon4 {background-color: #ddd; text-color: #000}
.defcon5 {background-color: #f7f7f7; text-color: #000}
.probe {padding-left: 40pt}
.detector {padding-left: 65pt}
h2 {padding-left: 20pt}
h3 {padding-left: 40pt}
h4 {padding-left: 60pt}
h2,h3,h4 {padding-top: 10px; padding-bottom: 10px}

/* Style the buttons that are used to open and close the accordion panel */
.accordion {
//  background-color: #eee;
  color: #444;
  cursor: pointer;
  padding: 18px;
  width: 100%;
  text-align: left;
  border: none;
  outline: none;
  transition: 0.4s;
  margin: 1pt;
}

/* Add a background color to the button if it is clicked on (add the .active class with JS), and when you move the mouse over it (hover) */
.active, .accordion:hover {
  background-color: #ccf;
}

/* Style the accordion panel. Note: hidden by default */
.panel {
  padding: 0 18px;
  background-color: white;
  display: none;
  overflow: hidden;
}
</style>
<title>garak report: {{reportfile}}</title>
<meta name="description" content="" />
</head>

<body>

<h1>garak run: {{reportfile}}</h1>
<button class="accordion">⚙️ view config</button>
<div style="border:solid black 1px; padding: 5px; margin: 5px" class="panel">
<h2>config</h2>
<p>filename: {{reportfile}}</p>
<p>garak version: {{garak_version}}</p>
<p>generator: {{model_type}}.{{model_name}}</p>
<p>started at: {{start_time}}</p>
<p>run config: {{setup}}</p>
<p>probe spec: {{probespec}}</p>
</div>

{%if model_name %}
<h2>results: {{model_type}} / {{model_name}}</h2>
{%else%}
<h2>results:</h2>
{%endif%}